open Misc
open Lambda
open Asttypes
open GccTypes

let compile_tailcalls = ref
    (try ignore (Sys.getenv "GCCTAIL"); true
    with Not_found -> false)

let rec struct_const cst cont =
  match cst with
  | Const_base(Const_int n) when n < 0 -> LDC 0 :: LDC (-n) :: SUB :: cont
  | Const_base(Const_int n) -> LDC n :: cont
  | Const_base(Const_char c) -> LDC (int_of_char c) :: cont
  | Const_base(Const_int32 n) -> LDC (Int32.to_int n) :: cont
  | Const_pointer n -> LDC n :: cont
  | Const_block(tag, arg) ->
      GccBuiltins.new_block tag
        (List.map struct_const arg) cont
  (* Not implemented *)
  | Const_base(Const_string _) -> assert false (* Not implemented *)
  | Const_immstring _ -> assert false (* Not implemented *)
  | Const_base(Const_float _) -> assert false (* Not implemented *)
  | Const_base(Const_int64 _) -> assert false (* Not implemented *)
  | Const_base(Const_nativeint _) -> assert false (* Not implemented *)
  | Const_float_array _ -> assert false (* Not implemented *)

type env = {
  ce_size: int;
  ce_frame: (int * int) Ident.tbl
}

let empty_env =
  { ce_size = 0; ce_frame = Ident.empty; }

let mk_label =
  let cpt = ref 0 in
  fun name ->
    incr cpt;
    L (Printf.sprintf "%s_%d" name !cpt)

let pop cont =
  let lbl = mk_label "pop" in
  ATOM :: TSEL (lbl, lbl) :: LABEL lbl :: cont

let get_var env id =
  try
    let frame, pos = Ident.find_same id env.ce_frame in
    (env.ce_size - frame, pos)
  with Not_found ->
    Format.eprintf "%a@." Ident.print id;
    fatal_error ("gccGccgen.get_var: var " ^ Ident.unique_name id)

let accu_id = Ident.create "accu"
let accu2_id = Ident.create "accu2"
let dup env cont =
  try
    let frame, pos = get_var env accu_id in
    ST (frame, pos) :: LD (frame, pos) :: LD (frame, pos) :: cont
  with Not_found ->
    fatal_error "gccGccgen.dup"

let jump lbl cont =
  LDC 0 :: TSEL (lbl, lbl) :: cont

let function_to_compile = ref []
let push_function ?id ?(toplevel = false) env kind params body =
  let name = match id with
    | None -> "anonymous"
    | Some id -> Ident.unique_name id in
  let real_lbl = mk_label ("fun_" ^ name ^ "_real") in
  let app_lbl = mk_label ("fun_" ^ name ^ "_app") in
  function_to_compile :=
    (name, real_lbl, app_lbl, env, kind, params, body, toplevel) ::
    !function_to_compile;
  real_lbl, app_lbl

let allocate id (pos, env) =
  (pos + 1,
   { env with ce_frame = Ident.add id (env.ce_size, pos) env.ce_frame })

let push_recenv funs (pos, env) =
  List.fold_left
    (fun env (id,_,_,_) -> allocate id env)
    (pos, { env with ce_size = env.ce_size + 1 })
    funs

let for_loop_registers = ref Ident.empty

let static_raise_label = ref []
let blocks = ref []
let push_static_catch i vars handler =
  let handler_lbl = mk_label (Printf.sprintf "static_catch%d_handler" i) in
  static_raise_label := (i, (handler_lbl, vars)) :: !static_raise_label;
  blocks := (LABEL handler_lbl ::  handler) :: !blocks

let pop_static_catch () =
  static_raise_label := List.tl !static_raise_label
let find_static_raise i =
  List.assoc i !static_raise_label

let rec comp_expr ?id ?toplevel ?(is_tailcall = false) env exp cont =
  match exp with
  | Lconst cst -> struct_const cst cont
  | Lvar id ->
      COMMENT (Printf.sprintf "get var %s" (Ident.unique_name id)) ::
      let frame, pos = get_var env id  in
      LD (frame, pos) :: cont
  | Llet(_, id, arg, body) ->
      let frame, pos = get_var env id  in
      comp_expr ~id env arg
        (COMMENT (Printf.sprintf "set var %s" (Ident.unique_name id)) ::
         ST (frame, pos) :: comp_expr ~is_tailcall env body cont)
  | Lprim(Pignore, [x])
  | Lprim(Pidentity, [x]) ->
      comp_expr ~is_tailcall env x cont (* I'm quite proud of this one, the hardest of all *)
  | Lprim(Pnegint, [x]) ->
      LDC 0 :: comp_expr env x (SUB :: cont)
  | Lprim(Paddint, [x;y]) ->
      comp_expr env x (comp_expr env y (ADD :: cont))
  | Lprim(Poffsetint i, [x]) ->
      comp_expr env x
        (if i = 0 then cont
         else if i < 0 then LDC (-i) :: SUB :: cont
         else LDC i :: ADD :: cont)
  | Lprim(Psubint, [x;y]) ->
      comp_expr env x (comp_expr env y (SUB :: cont))
  | Lprim(Pmulint, [x;y]) ->
      comp_expr env x (comp_expr env y (MUL :: cont))
  | Lprim(Pdivint, [x;y]) ->
      comp_expr env x (comp_expr env y (DIV :: cont))
  | Lprim(Psequand, [x;y]) ->
      let cont_lbl = mk_label "seqand_cont" in
      let true_lbl = mk_label "seqand_true" in
      let false_lbl = mk_label "seqand_false" in
      comp_expr env x @@
      LDC 0 :: CEQ :: TSEL (false_lbl, true_lbl) ::
      LABEL false_lbl :: LDC 0 :: (jump cont_lbl @@
      LABEL true_lbl :: (comp_expr ~is_tailcall env y @@
      LABEL cont_lbl :: cont))
  | Lprim(Psequor, [x;y]) ->
      let cont_lbl = mk_label "seqor_cont" in
      let true_lbl = mk_label "seqor_true" in
      let false_lbl = mk_label "seqor_false" in
      comp_expr env x @@
      LDC 0 :: CEQ :: TSEL (false_lbl, true_lbl) ::
      LABEL true_lbl :: LDC 1 :: (jump cont_lbl @@
      LABEL false_lbl :: (comp_expr ~is_tailcall env y @@
      LABEL cont_lbl :: cont))
  | Lprim(Pnot, [x]) -> LDC 1 :: comp_expr env x (SUB :: cont)
  | Lprim(Pintcomp Ceq, [x;y]) ->
      let (frame, pos) = get_var env accu_id in
      let (frame2, pos2) = get_var env accu2_id in
      let false_lbl = mk_label "ceq_false" in
      let ints_lbl = mk_label "ceq_ints" in
      let cont_lbl = mk_label "ceq_cont" in
      (comp_expr env x @@ ST (frame, pos) ::
      (comp_expr env y @@ ST (frame2, pos2) ::
      LD (frame, pos) :: ATOM ::
      LD (frame2, pos2) :: ATOM ::
      ADD :: LDC 2 :: CEQ :: TSEL (ints_lbl, false_lbl) ::
      LABEL false_lbl :: LDC 0 :: (jump cont_lbl @@
      LABEL ints_lbl :: LD (frame, pos) :: LD (frame2, pos2) :: CEQ ::
      LABEL cont_lbl :: cont)))
  | Lprim(Pintcomp Cneq, [x;y]) ->
      let (frame, pos) = get_var env accu_id in
      let (frame2, pos2) = get_var env accu2_id in
      let true_lbl = mk_label "cneq_true" in
      let ints_lbl = mk_label "cneq_ints" in
      let cont_lbl = mk_label "cneq_cont" in
      (comp_expr env x @@ ST (frame, pos) ::
      (comp_expr env y @@ ST (frame2, pos2) ::
      LD (frame, pos) :: ATOM ::
      LD (frame2, pos2) :: ATOM ::
      ADD :: LDC 2 :: CEQ :: TSEL (ints_lbl, true_lbl) ::
      LABEL true_lbl :: LDC 1 :: (jump cont_lbl @@
      LABEL ints_lbl ::
      LDC 1 :: LD (frame, pos) :: LD (frame2, pos2) :: CEQ :: SUB ::
      LABEL cont_lbl :: cont)))
  | Lprim(Pintcomp Cgt, [x;y]) ->
      comp_expr env x (comp_expr env y (CGT :: cont))
  | Lprim(Pintcomp Cle, [x;y]) ->
      LDC 1 :: comp_expr env x (comp_expr env y (CGT :: SUB :: cont))
  | Lprim(Pintcomp Cge, [x;y]) ->
      comp_expr env x (comp_expr env y (CGTE :: cont))
  | Lprim(Pintcomp Clt, [x;y]) ->
      LDC 1 :: comp_expr env x (comp_expr env y (CGTE :: SUB :: cont))
  | Lprim( Pmakeblock (tag,_), arg) ->
      GccBuiltins.new_block tag (List.map (comp_expr env) arg) cont
  | Lprim (Pfield i, [b]) ->
      GccBuiltins.get_field (comp_expr env b []) i cont
  | Lprim (Psetfield (i,_), [b;x]) ->
      GccBuiltins.set_field
        (comp_expr env b []) i
        (comp_expr env x []) cont
  | Lprim(Praise, [v]) ->
     STOP :: cont

  | Lprim(Pmakearray _ (* Yahoo! no float array *), arg) ->
      GccBuiltins.new_block 0 (List.map (comp_expr env) arg) cont
  | Lprim (Parraylength _, [a]) ->
      GccBuiltins.get_size (comp_expr env a []) cont
  | Lprim ((Parrayrefu _ | Parrayrefs _), [a; off]) ->
      (* Please read: "Bound checking for dummies!" *)
      GccBuiltins.get_dyn_field (comp_expr env a []) (comp_expr env off []) cont
  | Lprim ((Parraysetu _ | Parraysets _), [a; off; v]) ->
      (* Please read: "Bound checking for dummies!" *)
      GccBuiltins.set_dyn_field
        (comp_expr env a []) (comp_expr env off []) (comp_expr env v []) cont

  | Lifthenelse(lcond, lif, lelse) ->
      let cont_lbl = mk_label "if_cont" in
      let one_lbl = mk_label "if_one" in
      let block_lbl = mk_label "if_block" in
      let zero_lbl = mk_label "if_zero" in
      let int_lbl = mk_label "if_int" in
      (comp_expr env lcond @@
      (dup env @@
       ATOM :: TSEL (int_lbl, block_lbl) ::
       LABEL int_lbl :: TSEL (one_lbl, zero_lbl) ::
       LABEL zero_lbl :: (comp_expr ~is_tailcall env lelse @@ (jump cont_lbl @@
       LABEL block_lbl :: (pop @@ LABEL one_lbl :: (comp_expr ~is_tailcall env lif @@
       LABEL cont_lbl :: cont))))))
  | Lapply(lfun, largs, _) ->
      List.fold_left
        (fun cont lam -> comp_expr env lam cont)
        (LDC (List.length largs) :: comp_expr env lfun (
             (if !compile_tailcalls && is_tailcall then
                TAP 1 else AP 1 ) :: cont))
        largs
  | Lfunction(kind, params, body) ->
      let real_lbl, app_lbl =
        push_function ?id ?toplevel env kind params body in
      LDF real_lbl :: DUM 2 :: LDF app_lbl ::
      LDF (GccBuiltins.lbl_y 2):: RAP 2 :: cont
  | Lprim(Pisout, [ hi ; arg ]) ->
      let frame, pos = get_var env accu_id in
      let cont =
        LDC 0 :: LD  (frame, pos) :: CGT ::
        LD  (frame, pos) :: comp_expr env hi (CGT :: ADD :: cont)
      in
      comp_expr env arg (ST (frame, pos) :: cont)
  | Lprim(Pccall { Primitive.prim_name = "percent_atom"}, [arg]) ->
      comp_expr env arg (ATOM :: cont)
  | Lprim(Pisint, [arg]) ->
      comp_expr env arg (ATOM :: cont)
  | Lprim(Pccall { Primitive.prim_name = "percent_debug"}, [arg]) ->
      comp_expr env arg (DBUG :: LDC 0 :: cont)
  | Lprim(Pccall { Primitive.prim_name = "percent_car"}, [arg]) ->
      comp_expr env arg (CAR :: cont)
  | Lprim(Pccall { Primitive.prim_name = "percent_cdr"}, [arg]) ->
      comp_expr env arg (CDR :: cont)
  | Lprim(Pccall { Primitive.prim_name = "percent_cons"}, [arg1; arg2]) ->
      comp_expr env arg1 (comp_expr env arg2 (CONS :: cont))
  | Lprim (Pandint, [arg1; arg2]) ->
      let (frame, pos) = get_var env accu_id in
      let (frame2, pos2) = get_var env accu2_id in
      (comp_expr env arg1 @@ ST (frame, pos) ::
      (comp_expr env arg2 @@ ST (frame2, pos2) ::
       GccBuiltins.bit_compare_8values GccBuiltins.BcAnd
         (LD (frame, pos)) (LD (frame2, pos2)) cont))
  | Lprim (Porint, [arg1; arg2]) ->
      let (frame, pos) = get_var env accu_id in
      let (frame2, pos2) = get_var env accu2_id in
      (comp_expr env arg1 @@ ST (frame, pos) ::
      (comp_expr env arg2 @@ ST (frame2, pos2) ::
       GccBuiltins.bit_compare_8values GccBuiltins.BcOr
         (LD (frame, pos)) (LD (frame2, pos2)) cont))
  | Lprim (Pxorint, [arg1; arg2]) ->
      let (frame, pos) = get_var env accu_id in
      let (frame2, pos2) = get_var env accu2_id in
      (comp_expr env arg1 @@ ST (frame, pos) ::
      (comp_expr env arg2 @@ ST (frame2, pos2) ::
       GccBuiltins.bit_compare_8values GccBuiltins.BcXor
         (LD (frame, pos)) (LD (frame2, pos2)) cont))
  | Lprim(Pccall descr, largs) ->
      let name = descr.Primitive.prim_name in
      LDC 0 :: LDC 0 :: LDC 0 :: LDC 0
      :: List.fold_right
        (comp_expr env)
        largs
        (LDF (L name) :: AP (List.length largs + 4) :: cont)
  | Lswitch(larg, sw) ->
      let blocks = ref [] in
      let push_block name data =
        let lbl = mk_label name in
        blocks := (LABEL lbl :: data) :: !blocks;
        lbl in
      let cont_lbl = mk_label "switch_cont" in
      let (frame, pos) = get_var env accu_id in
      let default_lbl = match sw.sw_failaction with
        | Some l -> push_block "default" (comp_expr ~is_tailcall env l (jump cont_lbl []))
        | None -> push_block "match_failure" [STOP]
      in
      let rec comp_switch = function
        | [] -> [ TSEL (default_lbl, default_lbl) ]
        | [ v, l ] ->
            let found_lbl =
              push_block ("case_" ^ string_of_int v)
                (comp_expr ~is_tailcall env l (jump cont_lbl [])) in
            [ LDC v ; CEQ ; TSEL (found_lbl, default_lbl) ]
        | (v, l) :: tl ->
            let found_lbl =
              push_block
                ("case_" ^ string_of_int v)
                (pop (comp_expr ~is_tailcall env l (jump cont_lbl []))) in
            let next_lbl =
              push_block ("case_pas_" ^ string_of_int v) (comp_switch tl) in
            dup env [ LDC v ; CEQ ; TSEL (found_lbl, next_lbl) ]
      in
      let block_lbl =
        push_block "if_block"
          (GccBuiltins.get_tag [LD (frame, pos)] @@ comp_switch sw.sw_blocks) in
      let const_lbl =
          push_block "if_const"
          (LD (frame, pos) :: comp_switch sw.sw_consts) in
      comp_expr env larg @@
      ST (frame, pos) ::
      LD (frame, pos) ::
      ATOM ::
      TSEL (const_lbl, block_lbl) ::
      List.flatten !blocks @
      LABEL cont_lbl ::
      cont
  | Lsequence(l1, l2) ->
     comp_expr env l1 (pop (comp_expr env l2 cont))
  | Lletrec(decls, body)
    when List.for_all
        (function (_, Lfunction(_,_,_)) -> true | _ -> false) decls ->
      let n = List.length decls in
      let funs =
        List.map
          (function (id, Lfunction(k,a,b)) -> (id, k, a, b)
                    | _ -> assert false) decls in
      let _, recenv = push_recenv funs (0, env) in
      let labels =
        List.map
          (fun (id, kind, params, body) ->
             push_function ~id recenv kind params body)
          funs in
      DUM n ::
      List.flatten
        (List.map
           (fun (real_lbl, app_lbl) ->
              [LDF real_lbl; DUM 2; LDF app_lbl ;
               LDF (GccBuiltins.lbl_y 2); RAP 2])
           labels) @
      LDF (GccBuiltins.lbl_z n) ::
      RAP n ::
      List.rev_map
        (fun (id,_,_,_) ->
           let frame, pos = get_var env id in
           ST (frame, pos)) funs @
      comp_expr env body cont
  | Lwhile(lcond, lbody) ->
      let start_lbl = mk_label "while_start"  in
      let test_lbl = mk_label "while_test"  in
      let exit_lbl = mk_label "while_exit"  in
      jump test_lbl @@
      LABEL start_lbl :: (comp_expr env lbody @@ (pop @@
      LABEL test_lbl :: (comp_expr env lcond @@ TSEL (start_lbl, exit_lbl) ::
      LABEL exit_lbl :: LDC 0 :: cont)))
  | Lfor(param, s, e, dir, body) ->
      let frame, pos = get_var env param in
      let s_id, e_id = Ident.find_same param !for_loop_registers in
      let e_frame, e_pos = get_var env e_id in
      let s_frame, s_pos = get_var env s_id in
      let start_lbl = mk_label "for_start"  in
      let test_lbl = mk_label "for_test"  in
      let exit_lbl = mk_label "for_exit"  in
      let cont =
        [ LD (frame, pos) ;
          LDC 1 ;
          (match dir with Upto -> ADD | Downto -> SUB) ;
          ST (frame, pos) ;
          LABEL test_lbl ]
        @ (match dir with Upto -> [ LD (frame, pos) ; LD (e_frame, e_pos) ]
                        | Downto -> [ LD (e_frame, e_pos) ; LD (frame, pos) ])
        @ [ CGT ; TSEL (exit_lbl, start_lbl) ]
        @ (LABEL exit_lbl :: LDC 0 :: cont) in
      let cont =
        jump test_lbl (LABEL start_lbl :: comp_expr env body (pop cont)) in
      let cont =
        LD (s_frame, s_pos)
        :: ST (frame, pos) :: cont in
      let cont = comp_expr env e (ST (e_frame, e_pos) :: cont) in
      comp_expr env s (ST (s_frame, s_pos) :: cont)
  | Lassign (id, expr) ->
      let frame, pos = get_var env id  in
      (comp_expr env expr @@
       COMMENT (Printf.sprintf "assign var %s" (Ident.unique_name id)) ::
       ST (frame, pos) :: LDC 0 :: cont)
  | Lstaticraise (i, ls) ->
      let (handler_lbl, vars) = find_static_raise i in
      List.fold_right2
        (fun var exp cont ->
           let frame, pos = get_var env var in
           comp_expr env exp (ST (frame, pos) :: cont))
        vars
        ls
        (jump handler_lbl cont)
  | Lstaticcatch(lbody, (i, vars), lhandler) ->
      let cont_lbl =
        mk_label (Printf.sprintf "static_catch%d_continuation" i) in
      push_static_catch i vars (comp_expr ~is_tailcall env lhandler (jump cont_lbl []));
      let res = comp_expr ~is_tailcall env lbody (LABEL cont_lbl :: cont) in
      pop_static_catch ();
      res

  (* Not implemented *)
  | Lletrec(id_arg_list, body) -> assert false (* Not implemented *)
  | Lprim(prim, largs) ->
      Format.eprintf "<<<<Unknown primitive: %a>>>>\n%!"
        Printlambda.primitive prim ;
      assert false (* Not implemented *)
  | Ltrywith(lbody, param, lhandler) -> assert false (* Not implemented *)
  | Lsend (k, met, obj, largs, _) -> assert false (* Not implemented *)
  | Levent(expr, ev) -> assert false (* Not implemented *)
  | Lifused(id, expr) -> assert false (* Not implemented *)

let merge (pos1, env1) (pos2, env2) =
  assert (env1.ce_size = env2.ce_size);
  (max pos1 pos2,
   { ce_size = env1.ce_size;
     ce_frame = Ident.fold_all Ident.add env1.ce_frame env2.ce_frame })

let rec allocate_frame env exp =
  match exp with
  | Lconst _ | Lvar _ -> env
  | Llet(_, id, arg, body) ->
      let arg_env = allocate_frame env arg in
      (* Damned! *)
      (* let body_env = allocate_frame (allocate id env) body in *)
      (* merge arg_env body_env *)
      allocate_frame (allocate id arg_env) body
  | Lprim(_, largs) -> allocate_frames env largs
  | Lifthenelse(lcond, lif, lelse) -> allocate_frames env [lcond; lif; lelse]
  | Lfunction(kind, params, body) -> env
  | Lapply(lfun, largs, _) -> allocate_frames env (lfun :: largs)
  | Lswitch(larg, sw) ->
      List.fold_left allocate_frame env
        (larg :: (List.split sw.sw_consts |> snd)
         @ (List.split sw.sw_blocks |> snd)
         @ (match sw.sw_failaction with None -> [] | Some l -> [l]))
  | Lsequence(l1, l2) -> allocate_frames env [l1; l2]
  | Lletrec(id_arg_list, body) ->
      let rec_env =
        List.fold_left
          (fun env (id, arg) -> allocate id env)
          env id_arg_list in
      allocate_frames rec_env (body :: List.map snd id_arg_list)
  | Lwhile(lcond, lbody) ->
      allocate_frames env [ lcond ; lbody ]
  | Lfor(param, lo, hi, dir, body) ->
      let env = allocate param env in
      let lo_id = (Ident.(create (name param ^ "_lo"))) in
      let hi_id = (Ident.(create (name param ^ "_hi"))) in
      for_loop_registers := Ident.add param (lo_id, hi_id) !for_loop_registers ;
      let env = allocate lo_id env in
      let env = allocate hi_id env in
      allocate_frames env [ lo ; hi ; body ]
  | Lassign(_, expr) -> allocate_frame env expr
  | Lstaticraise (i, ls) -> allocate_frames env ls
  | Lstaticcatch(lbody, (i, vars), lhandler) ->
      let body_env = allocate_frame env lbody in
      (* Damned! *)
      (* let handler_env = *)
        (* allocate_frame (List.fold_right allocate vars env) lhandler in *)
      (* merge body_env handler_env *)
      allocate_frame (List.fold_right allocate vars body_env) lhandler

  (* Not implemented *)
  | Ltrywith(lbody, param, lhandler) -> assert false (* Not implemented *)
  | Lsend (k, met, obj, largs, _) -> assert false (* Not implemented *)
  | Levent(expr, ev) -> assert false (* Not implemented *)
  | Lifused(id, expr) -> assert false (* Not implemented *)

and allocate_frames env lams =
  (* Damned! *)
  (* List.fold_left (fun renv l -> merge (allocate_frame env l) renv) env lams *)
  List.fold_left (fun env l -> allocate_frame env l) env lams


let rec build_partial_call name n i =
  let next_lbl = mk_label (Printf.sprintf "fun_%s_App_next_%d" name i) in
  let our_lbl = mk_label (Printf.sprintf "fun_%s_App_our_%d" name i) in
  if i = 0 then [] else
    [ LD (0, 0) ;
      LDC i ;
      CEQ ;
      TSEL (our_lbl, next_lbl) ;
      LABEL our_lbl ;
      LD (1, 1) (* Nous-même ! *) ;
      DUM (i+2) ;
      LDF (GccBuiltins.lbl_restart i) ;
      LDF (GccBuiltins.lbl_y (i+2));
      RAP (i+2) ;
      RTN ;
      LABEL next_lbl ] @ build_partial_call name n (i-1)

let rec dummy i = if i <= 0 then [] else LDC 0 :: dummy (pred i)

let compile_function (name, real_lbl, app_lbl, env,
                      kind, params, body, toplevel) =
  let env = { env with ce_size = env.ce_size + 1 } in
  let n, env = List.fold_right allocate params (0, env) in
  let size, env =
    if toplevel then
      allocate accu_id @@ allocate accu2_id @@ allocate_frame (n, env) body
    else
      allocate_frame (n, env) body
  in
  (* Printf.eprintf "ENV for %S (%d)\n%!" name env.ce_size; *)
  (* Ident.iter (fun id (frame, pos) -> *)
      (* Printf.eprintf " - %s (%d, %d)\n%!" (Ident.unique_name id) frame pos) *)
    (* env.ce_frame; *)
  let over_or_return =
    let return_lbl = mk_label ("fun_" ^ name ^ "_RETURN") in
    let over_lbl = mk_label ("fun_" ^ name ^ "_OVER") in
    [LD (0, size); LDC 0; CEQ; TSEL (return_lbl, over_lbl);
     LABEL return_lbl; RTN;
     LABEL over_lbl;
     LD (0, size);
     CONS;
     ST (0, size);
     LD (0, size);
     CDR;
     LD (0, size);
     CAR;
     TAP 1
    ] in
  let instrs = comp_expr
      ~is_tailcall:true
      env body over_or_return in
  let over_or_total_lbl = mk_label ("fun_" ^ name ^ "_over_or_total") in
  let partial_lbl = mk_label ("fun_" ^ name ^ "_partial") in
  blocks := (LABEL real_lbl :: instrs) :: !blocks;
  let app_instrs =
    [LD (0, 0);
     LDC n;
     CGTE;
     TSEL (over_or_total_lbl, partial_lbl);
     LABEL over_or_total_lbl] @
     dummy (size - n) @
    [LD(0, 0);
     LDC n;
     SUB;
     LD (1, 0); (* La vraie cloture ! *)
     TAP (size+1);
     LABEL partial_lbl;
    ] @ build_partial_call name n (n-1)
  in
  blocks := (LABEL app_lbl :: app_instrs) :: !blocks


let compile machine lam =
  Printf.eprintf "Compile with tailcall %b\n%!" !compile_tailcalls;
  let size, env = allocate_frame (0, empty_env) lam in
  assert (size = 0);
  let main_instr =
    comp_expr
      ~id:(Ident.create "entry")
      ~toplevel:true
      ~is_tailcall:true
      env lam [] in
  while !function_to_compile <> [] do
    let f = List.hd !function_to_compile in
    function_to_compile := List.tl !function_to_compile;
    compile_function f
  done;
  let instrs =
    (if machine then
       ([LD (0, 1); LD (0, 0); LDC 2] @ main_instr @ [ TAP 1 ])
     else
       ([LDC 0; LDC 0; LDC 2] @ main_instr @ [ TAP 1 ]))
    :: GccBuiltins.builtins ()
    :: !blocks
  in
  let instrs = List.flatten instrs in
  Array.of_list instrs
