
type state = {
  mutable direction: direction;
  (* locations: location list; *)
  fruit_position: (int * int) option;
  mutable utc: int;
  mutable fright_mode: int; (* remaining tick. Zero = normal mode *)
  mutable remaining_pill: int;
  mutable remaining_power_pill: int;
}


let first_fruit_utc = 127 * 200
let first_fruit_end_utc = 127 * 280
let second_fruit_utc = 127 * 400
let second_fruit_end_utc = 127 * 480

let fright_mode_duration = 20 * 127

(** Constants *)

(* A group of adjacent pill is associated with a 'level':

     min (group_level_limit_hi,
          group_level_limit_lo + N * pill_level + M * power_pill_level)

   Pills are marked with a level of [pill_deliciouness].
   Their neighbours are marked with a level :

     deliciousness - deliciousness / group_level

   With 'deliciousness = 2**30' and 'group_level = 100', the group
   is visible 1700 cells away.

   The deliciousness_bonus is used to maks small components visible
   from their surroundings so that the lambda man won't miss an easy
    small group catch.
*)

let pill_deliciouness = 1073741823 (* 2 ** 30 *)
(* let pill_deliciouness = 1000 (\* 2 ** 30 *\) (* for lol map *)*)

let group_level_limit_lo = 100
let power_pill_level = 60
let pill_level = 20
let group_level_limit_hi = 500

let ghost_danger_level = 2

let fruits_safe_belt = 5

let pill_deliciouness_bonus = 7


(** Helpers *)

let debug_array m =
  let h = array_length m in
  let w = array_length m.(0) in
  let map = ref (value 0) in
  for y = h - 1 downto 0 do
    let row = ref (value 0) in
    for x = w - 1 downto 0 do
      row := cons (value (m.(y).(x))) !row
    done ;
    map := cons !row !map
  done ;
  debug (cons (value 1665_1665) !map)

let array_incr m m' =
  let h = array_length m in
  let w = array_length m.(0) in
  let map = ref (value 0) in
  for y = h - 1 downto 0 do
    for x = w - 1 downto 0 do
      m.(y).(x) <- m.(y).(x) + m'.(y).(x)
    done ;
  done

let create_data ~w ~h v = array_init h (fun _ -> array_create w v)

(* Insert element in a sorted association list. *)
let rec insert (level: int) contents = function
  | [] -> [ (level, contents) ]
  | (level', contents') as hd :: tl as l ->
      if level' = level then
        (level, contents @ contents') :: tl
      else if level' > level then
        hd :: insert level contents tl
      else
        (level, contents) :: l

(* Propagates 'deliciousness/dangerness'. *)
let rec propagate map data elts v limit step =
  let do_step () =
    let neighs =
      list_fold_left
        (fun acc (x, y) ->
           if v <= data.(y).(x) then acc else
           let test x y neighs =
             if data.(y).(x) < v && map.(y).(x) <> Wall
             then (x, y) :: neighs
             else neighs in
           data.(y).(x) <- v; (* mark points *)
           test x (y + 1) @@ test x (y - 1) @@
           test (x + 1) y @@ test (x - 1) y @@ acc)
        [] elts in
    let new_v = step v in
    if neighs <> [] && v <> new_v && new_v > 0 then
      propagate map data neighs new_v limit step
  in
  match limit with
  | None -> do_step ()
  | Some l -> if v > l then do_step ()

(** A "component" is a group of adjacent cell. *)
type component = {
  level: int;
  elts: (int * int) list;
  neighs: (int * int) list;
}

let kind_level = function
  | Pill ->  pill_level
  | Power_pill -> power_pill_level
  | _ -> halt ()

let pill_components world =
  let map = world.map in
  let h = array_length map in
  let w = array_length map.(0) in
  let { x = sx ; y = sy } = world.lambda_man.location in
  let marks = create_data ~w ~h false in
  let mark x y = marks.(y).(x) <- true in
  let worklist = ref [ (sx, sy) ] in
  let components = ref [] in
  let rec complete elts neighs level x y =
    match map.(y).(x) with
    | Wall -> ()
    | Pill | Power_pill as kind ->
        level := !level + kind_level kind;
        if not (marks.(y).(x)) then begin
          mark x y ;
          elts := (x, y) :: !elts ;
          complete elts neighs level (x + 1) y ;
          complete elts neighs level (x - 1) y ;
          complete elts neighs level x (y + 1) ;
          complete elts neighs level x (y - 1)
        end
    | _ -> neighs := (x, y) :: !neighs
  in
  let rec search x y =
    if not (marks.(y).(x)) then
      match map.(y).(x) with
      | Wall -> ()
      | Pill | Power_pill -> collect x y
      | _ ->
          mark x y ;
          search (x + 1) y ; search (x - 1) y ;
          search x (y + 1) ; search x (y - 1)
  and collect x y =
    let elts = ref []
    and neighs = ref []
    and level = ref group_level_limit_lo in
    complete elts neighs level x y ;
    let level = min !level group_level_limit_hi in
    components := { level; elts = !elts; neighs = !neighs } :: !components ;
    worklist := !neighs @ !worklist
  in
  while !worklist <> [] do
    let iteration = !worklist in
    worklist := [] ;
    list_iter (fun (x, y) -> search x y) iteration
  done ;
  !components

(** Main step function *)

let ai_step world state =

  let map = world.map in
  let h = array_length map in
  let w = array_length map.(0) in
  let { x = sx ; y = sy } = world.lambda_man.location in
  let { utc } = state in
  let fright_mode = state.fright_mode > 0 in

  (** Compute gradiant of interest! *)
  let delicious =
    let decr_deliciousnness decr_rate v = v - v / decr_rate in
    let delicious = create_data ~w ~h 0 in
    let delicious_bonus = create_data ~w ~h 0 in
    let components = pill_components world in
    (* compute the initial max-joined gradient *)
    list_iter
      (fun c ->
         propagate map delicious c.elts
           pill_deliciouness None (decr_deliciousnness c.level))
      components;
    (* add-join a homogeneous bonus to the max-joined gradient
       to avoid missing small groups when passing by them *)
    list_iter
      (fun c ->
         propagate map delicious_bonus c.elts
           pill_deliciouness_bonus (Some 0) pred)
      components;
    array_incr delicious delicious_bonus ;
    delicious
  in

  (** Compute the distance to fruit *)
  let fruits =
    let fruits = create_data ~w ~h 0 in
    (match state.fruit_position with
     | None -> ()
     | Some pos ->
         if utc < first_fruit_utc then
           (* max_step: nombre de pas max avant la disparition du fruits *)
           let max_step = (first_fruit_end_utc - utc) / 127 in
           propagate map fruits [pos] max_step None pred
         else if utc < second_fruit_utc then
           let max_step = (second_fruit_end_utc - utc) / 127 in
           propagate map fruits [pos] max_step None pred);
    fruits (* fruits.(sy).(sx) -> combien de pas avant que je ne puisse
                                  plus l'attraper *)
  in

  (** Compute the distance to ghosts *)
  let ghosts_max_level = 10 in
  let danger =
    let danger = create_data ~w ~h 0 in
    let ghosts_positions =
      list_iter
        (fun { location = { x ; y } ; direction ; vitality } ->
           if vitality <> Invisible then
             let w d p =
               let (x, y) = d p in
               map.(y).(x) = Wall in
             let walk u d l r =
               let rec go lvl ((x, y) as p) =
                 danger.(y).(x) <- max danger.(y).(x) lvl ;
                 if lvl > 0 then
                   if w u p && w l p && w r p then
                     go (lvl - 1) (d p)
                   else begin
                     if not (w u p) then go (lvl - 1) (u p) ;
                     if not (w l p) then go (lvl - 1) (l p) ;
                     if not (w r p) then go (lvl - 1) (r p)
                   end
               in go ghosts_max_level (x, y)
             in
             let u (x, y) = x, y - 1 and d (x, y) = x, y + 1
             and l (x, y) = x - 1, y and r (x, y) = x + 1, y in
             danger.(y).(x) <- ghosts_max_level ;
             match direction with
             | Up -> walk u d l r
             | Down -> walk d u r l
             | Left -> walk l r d u
             | Right -> walk r l u d)
        world.ghosts in
    debug_array danger ;
    danger
  in

  (** And eventually take a decision *)
  let eval_neighs x y eval =
    let test x y dir r =
      if map.(y).(x) <> Wall then list_insert (eval x y, dir) r else r in
    test x (y - 1) Up @@ test x (y + 1) Down @@
    test (x - 1) y Left @@ test (x + 1) y Right @@ []
  in
  let choose eval =
    match eval_neighs sx sy eval with
    | [] -> Up
    | [_, dir] -> dir
    | (v1, dir1) :: (v2, dir2) :: _ ->
        if v1 = v2 && dir1 = opp state.direction then dir2 else dir1 in
  let ghosts_distance = ghosts_max_level - danger.(sy).(sx) in
  let direction =
    if not fright_mode && ghosts_distance <= ghost_danger_level then
      (debug 1111;
       choose (fun x y -> (danger.(y).(x), -delicious.(y).(x))))
    else if fright_mode &&
            ghosts_distance < (state.fright_mode / 140) then
      (debug 2222;
       choose (fun x y -> (-danger.(y).(x))))
    else if fruits.(sy).(sx) > 0 &&
            ((world.fruit = None && fruits.(sy).(sx) - 80 < 0)
             || world.fruit <> None)
         (* Correct: mais dangeureux: on reste sur place en attendant.
           et en plus, dans ce mode, il faudrait fuir les pillules... *)
         (* || (utc < second_fruit_utc && state.remaining_pill < 8) *)
    then
      (debug 3333;
       choose (fun x y -> -(fruits.(y).(x))))
    else
      (debug 4444;
       choose (fun x y -> -(delicious.(y).(x))))
  in

  (** Compute new state and returns *)

  state.direction <- direction;
  let next_cell =
    match direction with
    | Up -> map.(sy-1).(sx)
    | Right -> map.(sy).(sx+1)
    | Down -> map.(sy+1).(sx)
    | Left -> map.(sy).(sx-1) in

  (* TODO if next pill = Power_pill -> wait for ghost... *)

  let eating =
    (* TODO: que ce passe t'il si l'on meurt ? *)
    next_cell = Pill || next_cell = Power_pill ||
    (next_cell = Fruit && world.fruit <> None)
  in
  let utc_incr = if eating then 137 else 127 in
  state.utc <- utc + utc_incr;
  if next_cell = Power_pill then
    ( state.fright_mode <- 20 * 127;
      state.remaining_power_pill <- state.remaining_power_pill - 1 )
  else if fright_mode then
    state.fright_mode <- max 0 (state.fright_mode - utc_incr);
  if next_cell = Pill then
    state.remaining_pill <- state.remaining_pill - 1;

  (state, direction)



(** Initialisation *)

let explore map =
  let fruit = ref None in
  let pill = ref 0 in
  let power_pill = ref 0 in
  let h = array_length map in
  let w = array_length map.(0) in
  for y = 0 to h - 1 do
    for x = 0 to w - 1 do
      match map.(y).(x) with
      | Fruit -> fruit := Some (x, y)
      | Pill -> incr pill
      | Power_pill -> incr power_pill
      | _ -> ()
    done
  done;
  !fruit, !pill, !power_pill

let ai_init world =
  let fruit_position, remaining_pill, remaining_power_pill = explore world.map in
  { direction = Down;
    (* locations = [ world.lambda_man.location ] *)
    utc = 0;
    fright_mode = 0;
    fruit_position;
    remaining_pill;
    remaining_power_pill;
  }
