type instr =
  | Jmp of rvalue
  | Set of lvalue * rvalue
  | Arith of arith_op * lvalue * rvalue
  | Cmp of cmp_op * int * rvalue * rvalue
  | Interrupt of int
  | Halt
  | Undefined
 and rvalue =
   | Reg of reg
   | Cst of int
   | Mem_at_reg of reg
   | Mem_at_cst of int
   | Pc
 and lvalue =
   | Set_reg of reg
   | Set_mem_at_reg of reg
   | Set_mem_at_cst of int
 and arith_op =
   | Add | Sub | Mul | Div
   | And | Or | Xor
 and cmp_op =
   | Lt | Eq  | Gt
 and reg = int
 and interrupt = int

type ghc =
    { data : int array ;
      code : instr array ;
      mutable pc : int ;
      registers : int array }

let trace ghc =
  Printf.eprintf
    "< PC: %d ; REGS: [ %d, %d, %d, %d, %d, %d, %d, %d ]>\n%!"
    ghc.pc
    ghc.registers.(0) ghc.registers.(1) ghc.registers.(2) ghc.registers.(3)
    ghc.registers.(4) ghc.registers.(5) ghc.registers.(6) ghc.registers.(7)

let create () =
  let data = Array.make 256 0 in
  let code = Array.make 256 Undefined in
  let pc = 0 in
  let registers = Array.make 8 0 in
  { data ; code ; pc ; registers }

let from_file fp =
  let ghc = create () in
  let text =
    let fp = open_in fp in
    let size = in_channel_length fp in
    let text = String.make (size + 1) '\000' in
    really_input fp text 0 size ;
    close_in fp ;
    text
  in
  let rec spaces acc p i =
    match String.unsafe_get text i with
    | ' ' | '\t' -> spaces acc (i + 1) (i + 1)
    | 'a'..'z' | 'A'..'Z' -> ident acc i i
    | '0'..'9' -> number acc i i
    | '[' -> spaces (`O :: acc) (i + 1) (i + 1)
    | ']' -> (* very nice design *)
       (match acc with
	| `W r :: `O :: tl -> spaces (`W' r :: tl) (i + 1) (i + 1)
	| `N n :: `O :: tl -> spaces (`N' n :: tl) (i + 1) (i + 1)
	| _ -> failwith "syntax error, unexpected ']'")
    | ',' -> spaces (`C :: acc) (i + 1) (i + 1)
    | '\n' -> spaces (`R :: acc) (i + 1) (i + 1)
    | ';' -> comment acc i i
    | '\000' -> List.rev (`R :: acc)
    | c -> failwith (Printf.sprintf "unexpected character %C" c)
  and word p i = `W (String.uppercase (String.sub text p (i - p)))
  and num p i = `N (int_of_string (String.sub text p (i - p)))
  and ident acc p i =
    match String.unsafe_get text i with
    | 'a'..'z' | 'A'..'Z' -> ident acc p (i + 1)
    | _ -> spaces (word p i :: acc) i i
  and number acc p i =
    match String.unsafe_get text i with
    | '0'..'9' -> number acc p (i + 1)
    | _ -> spaces (num p i :: acc) i i
  and comment acc p i =
    match String.unsafe_get text i with
    | '\n' | '\000' -> spaces acc i i
    | c -> comment acc p (i + 1)
  in
  let tokens = spaces [] 0 0 in
  let rec parse i toks =
    if i > 256 then
      failwith "too many instructions"
    else
      let regn n =
	if String.length n > 1 then
	  failwith (Printf.sprintf "bad register name %S" n)
	else
	  let r = Char.code n.[0] - Char.code 'A' in
	  if r < 0 || r > 7 then
	    failwith (Printf.sprintf "bad register name %S" n)
	  else r
      in
      let arith_op = function
	| "ADD" -> Add | "SUB" -> Sub | "MUL" -> Mul | "DIV" -> Div
	| "AND" -> And | "OR" -> Or | "XOR" -> Xor
	| _ -> assert false
      and cmp_op = function
	| "JLT" -> Lt | "JEQ" -> Eq | "JGT" -> Gt
	| _ -> assert false
      and rvalue = function
	| `W "PC" -> Pc
	| `W r -> Reg (regn r)
	| `N n -> Cst n
	| `W' r -> Mem_at_reg (regn r)
	| `N' n -> Mem_at_cst n
      and lvalue = function
	| `W r -> Set_reg (regn r)
	| `W' r -> Set_mem_at_reg (regn r)
	| `N' n -> Set_mem_at_cst n
      in
      match toks with
      | `W "MOV" :: `W "PC" :: `C :: `N cst :: `R :: tl ->
	 ghc.code.(i) <- Jmp (Cst cst) ;
	 parse (i + 1) tl
      | `W "MOV" :: `W "PC" :: `C :: `W "PC" :: `R :: tl ->
	 ghc.code.(i) <- Jmp Pc ;
	 parse (i + 1) tl
      | `W "MOV" :: `W "PC" :: `C :: `W reg :: `R :: tl ->
	 ghc.code.(i) <- Jmp (Reg (regn reg)) ;
	 parse (i + 1) tl
      | `W "MOV"
	:: (`W _ | `W' _ | `N' _ as dst) :: `C
 	:: (`W' _ | `N' _ | `W _ | `N _ as src) :: `R :: tl ->
	 ghc.code.(i) <- Set (lvalue dst, rvalue src) ;
	 parse (i + 1) tl
      | `W "INC" :: (`W _ | `W' _ | `N' _ as dst) :: `R :: tl ->
	 ghc.code.(i) <- Arith (Add, lvalue dst, Cst 1) ;
	 parse (i + 1) tl
      | `W "DEC" :: (`W _ | `W' _ | `N' _ as dst) :: `R :: tl ->
	 ghc.code.(i) <- Arith (Sub, lvalue dst, Cst 1) ;
	 parse (i + 1) tl
      | `W ("ADD" | "MUL" | "DIV" | "SUB" | "AND" | "OR" | "XOR" as op)
	:: (`W _ | `W' _ | `N' _ as dst) :: `C
 	:: (`W' _ | `N' _ | `W _ | `N _ as src) :: `R :: tl ->
	 ghc.code.(i) <- Arith (arith_op op, lvalue dst, rvalue src) ;
	 parse (i + 1) tl
      | `W ("JLT" | "JEQ" | "JGT" as op)
	:: `N addr :: `C
	:: (`W' _ | `N' _ | `W _ | `N _ as x) :: `C
	:: (`W' _ | `N' _ | `W _ | `N _ as y) :: `R :: tl ->
	 ghc.code.(i) <- Cmp (cmp_op op, addr, rvalue x, rvalue y) ;
	 parse (i + 1) tl
      | `W "INT" :: `N n :: `R :: tl ->
	 ghc.code.(i) <- Interrupt n ;
	 parse (i + 1) tl
      | `W "HLT" :: `R :: tl ->
	 ghc.code.(i) <- Halt ;
	 parse (i + 1) tl
      | `R :: tl ->
	 parse i tl
      | `O :: _ -> failwith (Printf.sprintf "syntax error, unexpected '['")
      | `C :: _ -> failwith (Printf.sprintf "syntax error, unexpected coma")
      | `N n :: _ -> failwith (Printf.sprintf "syntax error, unexpected %d" n)
      | `W n :: _ -> failwith (Printf.sprintf "syntax error, unexpected %s" n)
      | `N' n :: _ -> failwith (Printf.sprintf "syntax error, unexpected [%d]" n)
      | `W' n :: _ -> failwith (Printf.sprintf "syntax error, unexpected [%s]" n)
      | [] -> ()
  in
  parse 0 tokens ;
  ghc

(* printer *)
let print ppf { code } =
  let chr n = Char.chr (Char.code 'a' + n) in
  let rec print_instr = function
    | Jmp src ->
       Format.fprintf ppf "mov pc, %a\n" print_rvalue src
    | Set (dst, src) ->
       Format.fprintf ppf "mov %a, %a\n" print_lvalue dst print_rvalue src
    | Arith (Sub, dst, Cst 1) ->
       Format.fprintf ppf "dec %a\n" print_lvalue dst
    | Arith (Add, dst, Cst 1) ->
       Format.fprintf ppf "inc %a\n" print_lvalue dst
    | Arith (op, dst, src) ->
       Format.fprintf ppf "%a %a, %a\n" print_arith_op op print_lvalue dst print_rvalue src
    | Cmp (op, addr, x, y) ->
       Format.fprintf ppf "%a %d, %a, %a\n" print_cmp_op op addr print_rvalue x print_rvalue y
    | Interrupt i ->
       Format.fprintf ppf "int %d\n" i
    | Halt ->
       Format.fprintf ppf "hlt\n"
    | Undefined ->
       assert false
 and print_rvalue ppf = function
   | Reg n ->
      Format.fprintf ppf "%c" (chr n)
   | Cst n ->
      Format.fprintf ppf "%d" n
   | Mem_at_reg n ->
      Format.fprintf ppf "[%c]" (chr n)
   | Mem_at_cst n ->
      Format.fprintf ppf "[%d]" n
   | Pc ->
      Format.fprintf ppf "pc"
 and print_lvalue ppf = function
   | Set_reg n ->
      Format.fprintf ppf "%c" (chr n)
   | Set_mem_at_reg n ->
      Format.fprintf ppf "[%c]" (chr n)
   | Set_mem_at_cst n ->
      Format.fprintf ppf "[%d]" n
 and print_arith_op ppf = function
   | Add -> Format.fprintf ppf "add" 
   | Sub -> Format.fprintf ppf "sub" 
   | Mul -> Format.fprintf ppf "mul" 
   | Div -> Format.fprintf ppf "div" 
   | And -> Format.fprintf ppf "and" 
   | Or -> Format.fprintf ppf "or" 
   | Xor -> Format.fprintf ppf "xor" 
 and print_cmp_op ppf = function
   | Lt -> Format.fprintf ppf "jlt" 
   | Eq -> Format.fprintf ppf "jeq" 
   | Gt -> Format.fprintf ppf "jgt" 
  in
  let rec print_code i =
    if i >= Array.length code
    then (if i > 255 then Format.eprintf "too long... @."; ())
    else
      match code.(i) with
      | Undefined -> ()
      | instr -> print_instr instr ; print_code (i + 1)
  in print_code 0

(* world accessors *)
type interrupt_handlers =
    { get_fst_lambda_man_pos : unit -> int * int ;
      get_snd_lambda_man_pos : unit -> int * int ;
      get_ghost_index : unit -> int ;
      get_ghost_starting_pos : int -> int * int ;
      get_ghost_current_pos : int -> int * int ;
      get_ghost_info : int -> World.ghost_vitality * World.direction ;
      get_map_at : int -> int -> World.cell ;
      get_registers : unit -> int * int array }

open World

let encode_direction =
  function Up -> 0 | Right -> 1 | Down -> 2 | Left -> 3
let decode_direction =
  function 0 -> Up | 1 -> Right | 2 -> Down | 3 -> Left | _ -> failwith "decode_direction"

let encode_vitality =
  function Standard -> 0 | Fright_mode -> 1 | Invisible -> 2
let decode_vitality =
  function 0 -> Standard | 1 -> Fright_mode | 2 -> Invisible | _ -> failwith "decode_vitality"

let encode_cell = function
  | Empty -> 1 | Wall -> 0 | Pill -> 2
  | Power_pill -> 3 | Fruit -> 4 | Lambda_man -> 5 | Ghost -> 6

let decode_cell = function
  | 1 -> Empty | 0 -> Wall | 2 -> Pill | 3 -> Power_pill
  | 4 -> Fruit | 5 -> Lambda_man | 6 -> Ghost | _ -> failwith "decode_cell"

(* main runner *)
let interp ghc world index =
  let open World in
  let dir = ref world.ghosts.(index).direction in
  let rec interp i =
    if i = 1024 then ()
    else match ghc.code.(ghc.pc) with
	 | Halt -> ()
	 | instr ->
	    let pc = ghc.pc in
	    interp_instr instr ;
	    if ghc.pc = pc then ghc.pc <- ghc.pc + 1 ;
	    interp (i + 1)
  and interp_instr = function
    | Jmp src ->
       ghc.pc <- interp_rvalue src
    | Set (dst, src) ->
       interp_set_lvalue dst (interp_rvalue src)
    | Arith (op, dst, src) ->
       interp_set_lvalue dst (interp_arith_op op (interp_lvalue dst) (interp_rvalue src))
    | Cmp (op, addr, x, y) ->
       if interp_cmp_op op (interp_rvalue x) (interp_rvalue y) then
	 ghc.pc <- addr
    | Interrupt 0 ->
       dir := decode_direction ghc.registers.(0)
    | Interrupt 1 ->
       let { x ; y } = world.lambda_man.location in
       ghc.registers.(0) <- x ;
       ghc.registers.(1) <- y
    | Interrupt 2 ->
       failwith ("interrupt 2 not handled")
    | Interrupt 3 ->
       ghc.registers.(0) <- index ;
    | Interrupt 4 ->
       let index = ghc.registers.(0) in
       let { x ; y } = world.ghosts.(index).starting_location in
       ghc.registers.(0) <- x ;
       ghc.registers.(1) <- y
    | Interrupt 5 ->
       let index = ghc.registers.(0) in
       let { x ; y } = world.ghosts.(index).current_location in
       ghc.registers.(0) <- x ;
       ghc.registers.(1) <- y
    | Interrupt 6 ->
       let index = ghc.registers.(0) in
       let { vitality ; direction } = world.ghosts.(index) in
       ghc.registers.(0) <- encode_vitality vitality ;
       ghc.registers.(1) <- encode_direction direction
    | Interrupt 7 ->
       let x = ghc.registers.(0) in
       let y = ghc.registers.(1) in
       ghc.registers.(0) <- encode_cell world.map.(y).(x)
    | Interrupt 8 ->
       trace ghc
    | Interrupt i ->
       failwith (Printf.sprintf "unknown interrupt %d" i)
    | Halt ->
       assert false
    | Undefined ->
       failwith "jumped out of the GHC program"
 and interp_rvalue = function
   | Reg n -> ghc.registers.(n)
   | Cst n -> n
   | Mem_at_reg n -> ghc.data.(ghc.registers.(n))
   | Mem_at_cst n -> ghc.data.(n)
   | Pc -> ghc.pc
 and interp_lvalue = function
   | Set_reg n -> ghc.registers.(n)
   | Set_mem_at_reg n -> ghc.data.(ghc.registers.(n))
   | Set_mem_at_cst n -> ghc.data.(n)
 and interp_set_lvalue dst src = match dst with
   | Set_reg n -> ghc.registers.(n) <- src
   | Set_mem_at_reg n -> ghc.data.(ghc.registers.(n)) <- src
   | Set_mem_at_cst n -> ghc.data.(n) <- src
 and interp_arith_op op x y = match op with
   | Add -> (x + y) land 0xFF
   | Sub -> (x - y) land 0xFF
   | Mul -> (x * y) land 0xFF
   | Div -> (x / y) land 0xFF
   | And -> x land y
   | Or -> x lor y
   | Xor -> (x lxor y) land 0xFF
 and interp_cmp_op op x y = match op with
   | Lt -> x < y
   | Eq -> x = y
   | Gt -> x > y
  in
  ghc.pc <- 0 ;
  interp 0 ;
  !dir
