
type cell =
  | Empty
  | Wall
  | Pill
  | Power_pill
  | Fruit
  | Lambda_man
  | Ghost

type location = { x : int; y : int }

type direction = Up | Down | Left | Right

type lambda_man =
    { vitality : int;
      (** n = 0 : normal
          n > 0 : number of ticks with active power pill *)
      starting_location : location;
      location : location;
      direction : direction;
      remaining_lives : int;
      score : int }

type ghost_vitality =
  | Standard
  | Fright_mode
  | Invisible

type ghost =
    { vitality : ghost_vitality;
      current_location : location;
      starting_location : location;
      direction : direction }

type ghosts = ghost array

type world =
    { map : cell array array;
      lambda_man : lambda_man;
      ghosts : ghosts;
      fruit : int option }

let int i = GccTypes.TAG_INT i
let block (a, b) = GccTypes.TAG_CONS (a,b)

let cell_of_char = function
  | ' ' -> Empty
  | '#' -> Wall
  | '.' -> Pill
  | 'o' -> Power_pill
  | '%' -> Fruit
  | '\\' -> Lambda_man
  | '=' -> Ghost
  | c -> failwith (Printf.sprintf "cell_of_char: %c" c)

let char_of_cell = function
  | Empty -> ' '
  | Wall -> '#'
  | Pill -> '.'
  | Power_pill -> 'o'
  | Fruit -> '%'
  |  Lambda_man -> '\\'
  | Ghost -> '='

let cell_of_int = function
  | 0 -> Wall
  | 1 -> Empty
  | 2 -> Pill
  | 3 -> Power_pill
  | 4 -> Fruit
  | 5 -> Lambda_man
  | 6 -> Ghost
  | i -> failwith (Printf.sprintf "cell_of_int: %i" i)

let int_of_cell = function
  | Wall -> 0
  | Empty -> 1
  | Pill -> 2
  | Power_pill -> 3
  | Fruit -> 4
  | Lambda_man -> 5
  | Ghost -> 6

let rec value_list_of_list = function
  | [] -> int 0
  | h :: t -> block (h, value_list_of_list t)

let rec value_tuple_of_list = function
  | [] | [_] -> failwith "value_tuple_of_list"
  | [ a; b] -> block (a, b)
  | h :: t -> block (h, value_tuple_of_list t)

let value_of_cell c = int (int_of_cell c)

let value_of_map map =
  Array.map
    (fun a ->
     Array.map value_of_cell a
     |> Array.to_list
     |> value_list_of_list)
    map
  |> Array.to_list
  |> value_list_of_list

let value_of_location { x; y } =
  block (int x, int y)

let int_of_direction = function
  | Up -> 0
  | Right -> 1
  | Down -> 2
  | Left -> 3

let direction_of_int = function
  | 0 -> Some Up
  | 1 -> Some Right
  | 2 -> Some Down
  | 3 -> Some Left
  | _ -> None

let value_of_direction d = int (int_of_direction d)

let value_of_lambda_man (lambda_man:lambda_man) =
  value_tuple_of_list
    [ int lambda_man.vitality;
      value_of_location lambda_man.location;
      value_of_direction lambda_man.direction;
      int lambda_man.remaining_lives;
      int lambda_man.score ]

let int_of_ghost_vitality = function
  | Standard -> 0
  | Fright_mode -> 1
  | Invisible -> 2

let value_of_ghost_vitality v = int (int_of_ghost_vitality v)

let value_of_ghost (g:ghost) =
  value_tuple_of_list
    [value_of_ghost_vitality g.vitality;
     value_of_location g.current_location;
     value_of_direction g.direction]

let value_of_ghosts g =
  value_list_of_list
    (List.map value_of_ghost (Array.to_list g))

let value_of_fruit f =
  match f with
  | None -> int 0
  | Some i -> int i

let value_of_world world =
  value_tuple_of_list
    [value_of_map world.map;
     value_of_lambda_man world.lambda_man;
     value_of_ghosts world.ghosts;
     value_of_fruit world.fruit]


let from_file name =
  let text =
    let fp = open_in name in
    let rec input acc =
      match try Some (input_line fp) with End_of_file -> None with
      | Some line -> input (String.trim line :: acc)
      | None -> close_in fp ; Array.of_list (List.rev acc)
    in input []
  in
  let map, ghosts, lambda_man =
    let ghosts = ref [] in
    let lambda_man = ref None in
    let mlen = Array.length text in
    let amap = Array.make mlen [| |] in
    for y = 0 to mlen - 1 do
      let len = String.length text.(y) in
      let arr = Array.make len Wall in
      amap.(y) <- arr ;
      for x = 0 to len - 1 do
	let cell = cell_of_char text.(y).[x] in
	arr.(x) <- cell ;
	match cell with
	| Wall | Empty | Pill | Power_pill | Fruit -> ()
	| Lambda_man when !lambda_man <> None ->
	   failwith "MORE THAN ONE LAMBDA MAN, WTF DUDE"
	| Lambda_man ->
	   lambda_man := Some { vitality = 0 ;
			        starting_location = { x ; y } ;
				location = { x ; y } ;
				direction = Down ;
				remaining_lives = 3 ;
				score = 0 }
	| Ghost ->
	   let ghost = { vitality = Standard ;
			 starting_location = { x ; y } ;
			 current_location = { x ; y } ;
			 direction = Down } in
	   ghosts := ghost :: !ghosts
      done
    done ;
    amap, Array.of_list (List.rev !ghosts),
    match !lambda_man with
    | Some lambda_man -> lambda_man
    | None -> failwith "NO LAMBDA MAN, WTF DUDE"
  in { map ; ghosts ; lambda_man ; fruit = None }


let pill_count world =
  let r = ref 0 in
  for i = 0 to Array.length world.map - 1 do
    for j = 0 to Array.length world.map.(0) - 1 do
      match world.map.(i).(j) with
      | Pill -> incr r
      | _ -> ()
    done
  done;
  !r
